@@ -1,4 +1,4 @@
-/* Copyright (c) 2015 Nicira, Inc.
+/* Copyright (c) 2015, 2016 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,12 +18,21 @@
 
 #include <stdbool.h>
 
-struct controller_ctx;
 struct ovsdb_idl;
+struct ovsdb_idl_index;
+struct ovsdb_idl_txn;
 struct ovsrec_bridge;
+struct ovsrec_open_vswitch_table;
+struct sbrec_chassis;
+struct sbrec_chassis_table;
 
 void chassis_register_ovs_idl(struct ovsdb_idl *);
-void chassis_run(struct controller_ctx *, const char *chassis_id);
-bool chassis_cleanup(struct controller_ctx *, const char *chassis_id);
+const struct sbrec_chassis *chassis_run(
+    struct ovsdb_idl_txn *ovnsb_idl_txn,
+    struct ovsdb_idl_index *sbrec_chassis_by_name,
+    const struct ovsrec_open_vswitch_table *,
+    const char *chassis_id, const struct ovsrec_bridge *br_int);
+bool chassis_cleanup(struct ovsdb_idl_txn *ovnsb_idl_txn,
+                     const struct sbrec_chassis *);
 
 #endif /* ovn/chassis.h */
